unit SupraINI;
{
TSupraINI
Version 1.1 For all Delphi versions
Written By Eric A. Pettersen, A.K.A. Supralight Kaiilridge
Email: androndax@hotmail.com

Description:
A memory-based INI file reader/writer.
Meant to be an almost drop-in replacement for TIniFile that is
system-independent.
Should compile and run on Delphi 1-7. Hopefully all versions of Kylix.
Tested on: Delphi 1, Delphi 3, Delphi 6.
May have an error or two on Delphi 2...not sure if TObject has a destructor?

Main Differences:
#1. Can be created without a filename ('') to be initialized without loading a file.
#2. Must be told to Save its data, and exactly where. (does not default to Windir)
#3. Can load data from more than one file into one database.
    (does not check for entry dupes at this point - Returns whatever sorted first)
#4. Is case-sensitive with all Section and Key calls. (Entry and enTry are counted as different)
#5. Does NOT preserve Section or Key order when saving. Alphabetically sorted.
#6. Does NOT preserve comments when saving. (Discarded when loading)

Terms of use:
You may freely use this unit in any project you create, commercial, shareware,
or freeware with the following strings attatched:
Please inform me that you're using it and in what. :)
If you modify or improve upon it, please send me the modified unit.
If you really feel generous, a finished copy of whatever the program was. :P
Honestly, the third's a joke but if you want to...I won't complain. ;)

---Access Functions (outdated)---
function ReadString(Section, Key, vDefault : string) : string;
Reads out any value and returns it as a string. vDefault is returned
if it cannot find the entry.

function ReadBool(Section, Key : string) : boolean;
Reads a value and returns True or False according to these items:
Y, YES, TRUE, 1 = True
Anything else (Including no entry) = false
Not case-sensitive.

function WriteString(Section, Key, value : string) : integer;
Writes a value to the database. Modifies existing entry or creates
one if one doesn't exist, just like the standard Windows INI routines.

function WriteBool(Section, Key : string; value : boolean) : integer;
Writes a boolean value to the database. Modifies existing entry or creates
one if one doesn't exist, just like the standard Windows INI routines.

---Save/Load functions---
RawToFile(filename : string);
RawFromFile(filename : string);
Dumps raw database to specified file or loads raw database from that file.
Loading clears wheatever was in the database before.

function LoadFromStream(thestream : TStream): integer;
function LoadFromFile(filename : String): integer;
Load INI data from a file or stream, clearing what was already stored.

function AddFromStream(thestream : TStream): integer;
function AddFromFile(filename : String): integer;
Add INI data from a file or stream.

function SaveToStream(thestream : TStream): integer;
function SaveToFile(filename : String): integer;
Save reencoded INI data to a file or stream.
Clears the temporary input buffer. (Uses it as output buffer)

Current version function lacks:
Raw database and INI data *merge* functions.

}
interface
uses
 {Windows, Messages,} SysUtils, Classes;

 Type
 TSupraINI = class(TObject)
  private
   FAboutText : String;
   FMainData : TStringList; {Main database }
   FTempData : TStringList; {Temporary storage for I/O}
   function GetValue(str : string) : string; {Retrieves a value from the database}
   function GetSectionStart(section : string) : integer; {Returns the index of the start of a section}
   function GetSectionEnd(section : string; start : integer): integer; {Returns the index of the end of a section}
   function ReadTempData : integer; {Parse INI data in FTempData}
   function BuildINIData : integer; {Build valid INI file and place it in FTempData}
   function GetEntryPosition(section, Key : string) : integer;
  protected
   function GetCharPos(str : string; character : char) : integer; {Library-Type function}
  public
   constructor Create(filename : string);
   destructor Destroy; override;
   function Parse : integer; {Parse import buffer}
   {INI Load/Save functions
    Save creates a completely valid INI file.}
   procedure LoadFromStream(thestream : TStream);
   procedure LoadFromFile(filename : String);
   procedure AddFromStream(thestream : TStream);  {Add INI data to current database}
   procedure AddFromFile(filename : String);
   procedure SaveToStream(thestream : TStream);
   procedure SaveToFile(filename : String);
   {Raw database I/O functions (Save/Load the raw database)}
   procedure RawToFile(filename : string);
   procedure RawToStream(output : TStream);
   procedure RawFromFile(filename : string);
   procedure RawFromStream(input : TStream);
   {Management functions}
   procedure ClearDatabase;
   procedure EraseSection(Section: string);   {Erases a section from the database}
   procedure DeleteKey(Section, Key: string); {Delete a single key from the database}
   {Indexing Functions}
   procedure ReadSections(Strings: TStrings); {Retrieve the names of all sections in an INI file into a string list object. }
   procedure ReadSection (Section: string; Strings: TStrings); {Read all the key names in a section into a string list}
   procedure ReadSectionValues(const Section: string; Strings: TStrings);
             {Read all the values from a section's keys into a string list}
   {Data handling functions}
   function ReadString(Section, Key, Vdefault : string): string; {Read a string value }
   function ReadInteger(Section, Key : string; Vdefault : integer): integer; {Read an integer value}
   function ReadBool(Section, Key : string): boolean; {Read a Boolean value }
   function ReadFloat(Section, Key : string; Vdefault : extended): extended; {Read a Floating-point value}
   procedure WriteString(Section, Key, value : string); {Write a string value}
   procedure WriteInteger(Section, Key : string; value : integer); {Write an integer value}
   procedure WriteBool(Section, Key : string; value : boolean); {Write a boolean value}
   procedure WriteFloat(Section, Key : string; value : extended); {Write a floating-point value}
   {Properties}
   property About : String read FAboutText;
   property TempData : TStringList read FTempData; {Direct access to import cache}
  end;

implementation

const
{Internal copyright}
 copyrighttext = 'SupraINI Version 1.1 by Supralight';
{Database parsing and storage constants}
 SectionChar = '[';
 SectionEnd  = ']';
 ValueEquals = '=';
 CommentChar = ';';
 Delimiter   = ',';
{Boolean types for output}
 BoolOne     = '1';
 BoolZero    = '0';
{Other}
 StartDefault = 'SupraINI: No Entry (3946DFGHI378834SUPRA872385128INI7659283476)';


{
------------------------------------------------------------
--------------------Creation/Destruction--------------------
------------------------------------------------------------
}
constructor TSupraINI.Create(filename : string);
begin
FAboutText := copyrighttext;
FMainData := TStringList.Create;
FTempData := TStringList.Create;
if filename <> '' then LoadFromFile(filename);
end;

destructor TSupraINI.Destroy;
begin
ClearDatabase;
FMainData.Free;
FTempData.Free;
{$IFNDEF VER80} {note: Unknown if D2 TObject has destructor...Delphi 2=VER90}
inherited;
{$ENDIF}
end;

{
------------------------------------------------------------
----------------------Parsing Functions---------------------
------------------------------------------------------------
}
function TSupraINI.BuildINIData : integer;
{Builds a complete, valid INI file from the database, and
places it in the Temporary Data Buffer (FTempData) stringlist.}
var
 position : integer;
 working, Working2 : string;
 Section : string;
begin
Section := '';
{Builder loop}
For position := 0 to fMaindata.count-1 do
 begin
 {Get and Build/Ignore Section data}
 Working := FMainData.strings[position];
 Working2 := copy(working, 1, Pos(Delimiter,working)-1);
 delete(working, 1, Pos(Delimiter,working));
 if Working2 <> Section then {Section change, add a space then switch }
  begin                      {current section.}
  if position <> 0 then Ftempdata.Add('');
  FTempData.Add(SectionChar+Working2+SectionEnd);
  Section := Working2;
  end;
 {Build a Key/Value entry}
 Working2 := copy(working, 1, Pos(Delimiter,working)-1);
 Delete(working, 1, Pos(Delimiter,working));
 FtempData.add(working2+Valueequals+working);
 end;
Result := 0;
end;

function TSupraINI.ReadTempData : integer;
{Read and parse INI data from memory cache.
After reading, clears the cache.}
var
 test : integer;
 pos1 : integer;
 Section : string;
 temp, temp2, temp3 : string;
begin
{Begin loop through temporary data }
For pos1 := 0 to FTempData.count-1 do
 begin
 temp := FTempData[pos1];
 if length(temp) > 0 then  {Check length to make sure we don't try to parse}
  begin                    {an empty line, causing an access violation.}
  case temp[1] of
   SectionChar : begin {New Section discovered}
                 Delete(temp,1,1);
                 Delete(Temp,length(temp), length(temp));
                 Section := temp;
                 end;
   commentchar : begin {Comment discovered}
                 {Do nothing - ignore it.}
                 end;
   else  {Anything else, other than a blank line. Assume it's a value.}
   Temp2 := temp;         {Copy data in readiness for extraction}
   Temp3 := temp;
   Test := Pos(ValueEquals, temp);    {Check for the key/value separator}
   if test <> 0 then      {If the search didn't return 0, then}
    begin                 {extract both key and value, then add both}
    Delete(temp2,test,length(temp2)); {to the database along with the}
    Delete(temp3,1,test);             {section name.}
    FMainData.Add(section+delimiter+temp2+Delimiter+temp3);
    end;
   end;
  end;
 end;
FTempData.Clear; {empty temporary storage after parsing and conversion}
FMainData.Sort;  {Sort the main database table alphabetically}
result := 0;
end;

function TSupraINI.Parse : integer;
{Simple wrapper function for ReadTempData}
begin
result := ReadtempData;
end;

{
------------------------------------------------------------
-----------------------Data Retrieval-----------------------
------------------------------------------------------------
}
function TSupraINI.GetSectionStart(section : string) : integer;
{Get the start of a section in the database.
As they're alphabetically sorted, a single section will be all
together.}
var
 position : integer;
 temp : integer;
 found : boolean;
 finalposition : integer;
 temp2 : string;
 str : string;
begin
Found := false;
finalposition := -1;
str := Section;
for position := 0 to FMainData.Count-1 do {Iterate through database looking}
 begin                                    {for section }
 temp := pos(str, FMainData.Strings[position]);
 if (temp = 1) {If we get a positive, first check to make sure}
 then             {It's at the beginning of the line.}
  begin
  Temp2 := FMainData.Strings[position];
  if(temp2[length(str)+1] = Delimiter) then {Now check to make sure it's the}
   begin                              {complete field.}
   found := true;
   finalPosition := position;{Extract and clip value from other database fields}
   Break;
   end;
  end;
 end;
if found = false then Result := -1
 else Result := finalposition; {Return value data}
end;

function TSupraINI.GetSectionEnd(section : string; start : integer): integer;
var
 position : integer;
 temp : integer;
 found : boolean;
 finalposition : integer;
 temp2 : string;
 str : string;
begin
Found := false;
finalposition := -1;
str := Section;
for position := start to FMainData.Count-1 do {Iterate through database looking}
 begin                                    {for section }
 temp := pos(str, FMainData.Strings[position]);
 temp2 := FMainData.Strings[position];
 if (temp <> 1) and (temp2[length(str)+1] <> Delimiter)
 then  
  begin
  Finalposition := position-1;
  found := true;
  break;
  end;
 end;
if found = false then Result := -1
 else Result := finalposition; {Return value data }
end;

function TSupraINI.GetEntryPosition(section, Key : string) : integer;
{Returns the position of an entry in the database table
Return value of -1 indicates that the entry does not exist}
var
 position : integer;
 temp : integer;
 found : boolean;
 finalposition : integer;
 temp2 : string;
 str : string;
begin
Found := false;
finalposition := -1;
str := Section+Delimiter+Key;
for position := 0 to FMainData.Count-1 do{Iterate through database looking}
 begin                                    {for section and key}
 temp := pos(str, FMainData.Strings[position]);
 if (temp = 1) {If we get a positive, first check to make sure}
 then              {It's at the beginning of the line. }
  begin
  Temp2 := FMainData.Strings[position];
  if(temp2[length(str)+1] = Delimiter) then {Now check to make sure it's the}
   begin                              {complete field.}
   found := true;
   finalPosition := position;{Extract and clip value from other database fields}
   {Delete(CorrectStr, 1,length(str)+1);   }
   Break;
   end;
  end;
 end;
if found = false then Result := -1
 else Result := finalposition; {Return value data}
end;

function TSupraINI.GetValue(str : string) : string;
{Search for Str, and return everything after it, that's after the data
delimiter. If the next character is NOT the delimiter, this is not the
requested item, and keep searching.}
var
 position : integer;
 temp : integer;
 found : boolean;
 CorrectStr : string;
 temp2 : string;
begin
Found := false;
for position := 0 to FMainData.Count-1 do {Iterate through database looking}
 begin                                    {for section and key}
 temp := pos(str, FMainData.Strings[position]);
 if (temp = 1) {If we get a positive, first check to make sure}
 then              {It's at the beginning of the line.}
  begin
  Temp2 := FMainData.Strings[position];
  if(temp2[length(str)+1] = Delimiter) then {Now check to make sure it's the}
   begin                              {complete field.}
   found := true;
   CorrectStr := FMainData.Strings[position];{Extract and clip value from other database fields}
   Delete(CorrectStr, 1,length(str)+1);
   Break;
   end;
  end;
 end;
if found = false then Result := StartDefault {Return default if not found}
else
 begin
 Result := Correctstr; {Return value data }
 end;
end;

function TSupraINI.GetCharPos(str : string; character : char) : integer;
{Detect first occurrence of Character in a string and return the index.}
var
 posi, position: integer;
begin
position := 0;
for posi := 1 to length(str) do
 begin
 if str[posi]=character then
  begin
  position := posi;
  Break;
  end;
 end;
result := position;
end;

{
------------------------------------------------------------
---------------------Saving and Loading---------------------
------------------------------------------------------------
}
procedure TSupraINI.LoadFromStream(thestream : TStream);
{Load and parse INI data from a generic stream interface, clearing out
whatever's in the database right now.  }
begin
ClearDatabase;
FTempData.Clear;
FTempData.LoadFromStream(TheStream);
ReadTempData;
end;

procedure TSupraINI.LoadFromFile(filename : String);
{Load and parse INI data directly from a file, clearing out whatever's
in the database right now.}
begin
ClearDataBase;
FTempData.Clear;
FTempData.LoadFromFile(filename); {Load data into temporary line-based storage}
ReadTempData;
end;

procedure TSupraINI.AddFromStream(thestream : TStream);
{Load and parse INI data from a generic stream interface, adding to what's
in the current database.}
begin
FTempData.Clear;
FTempData.LoadFromStream(TheStream);
ReadTempData;
end;

procedure TSupraINI.AddFromFile(filename : String);
{Load and parse INI data directly from a file, adding to what's in the
current database.}
begin
FTempData.Clear;
FTempData.LoadFromFile(filename); {Load data into temporary line-based storage }
ReadTempData;
end;

procedure TSupraINI.SaveToStream(thestream : TStream);
{Save encoded INI data to a Delphi stream}
begin
FTempData.Clear;
BuildINIData;
FtempData.SavetoStream(thestream);
FTempData.clear;
end;

procedure TSupraINI.SaveToFile(filename : String);
{Save encoded INI data to a file}
begin
FTempData.Clear;
BuildINIData;
FtempData.SavetoFile(filename);
FTempData.clear;
end;

procedure TSupraINI.RawToFile(filename : string);
{Saves raw database to a file}
begin
FMainData.SaveToFile(filename);
end;

procedure TSupraINI.RawToStream(output : TStream);
begin
FmainData.SaveToStream(output);
end;

procedure TSupraINI.RawFromFile(filename : string);
begin
FMainData.LoadFromFile(filename);
end;

procedure TSupraINI.RawFromStream(input : TStream);
begin
FMainData.LoadFromStream(input);
end;

{
------------------------------------------------------------
---------------------Management Functions-------------------
------------------------------------------------------------
}
procedure TSupraINI.ClearDatabase;
begin
{Clear the database }
FMainData.clear;
FTempData.clear;
end;

procedure TSupraINI.EraseSection(Section: string);
{Erases a section from the database}
var
 linestart, lineend, count : integer;
begin
linestart := GetSectionstart(section);
lineend := GetSectionEnd(section,linestart);
for count := linestart to lineend do
 begin
 FMainData.Delete(linestart);
 end;
end;

procedure TSupraINI.DeleteKey(Section, Key: string);
{Delete a single key from the database}
var
 pos : integer;
begin
Pos := GetEntryPosition(section, Key);
FMaindata.Delete(pos);
end;

procedure TSupraINI.ReadSections(Strings: TStrings);
{Retrieve the names of all sections in an INI file into a string list object.
Modification of INI Builder function.}
var
 position : integer;
 working, Working2 : string;
 Section : string;
begin
Section := '';
For position := 0 to fMaindata.count-1 do
 begin
 {Get Section names}
 Working := FMainData.strings[position];
 Working2 := copy(working, 1, Pos(Delimiter,working)-1);
 delete(working, 1, Pos(Delimiter,working));
 if Working2 <> Section then {Section change, add the new section and switch}
  begin                      {current section.}
  Section := Working2;
  strings.Add(Working2);
  end;
 end;
end;

procedure TSupraINI.ReadSection(Section: string; Strings: TStrings);
{Read all the key names in a section into a string list
Modification of INI Builder function.}
var
 position : integer;
 working, Working2, working3 : string;
 CSection : string;
 havefound : boolean;
begin
cSection := ''; havefound := false;
For position := 0 to fMaindata.count-1 do
 begin
 {Get Section names}
 Working := FMainData.strings[position];
 Working2 := copy(working, 1, Pos(Delimiter,working)-1);
 delete(working, 1, Pos(Delimiter,working));
 if Working2 = Section then
  begin
  havefound := true;
  cSection := Working2;
  Working3 := copy(working, 1, Pos(Delimiter,working)-1);
  Delete(working, 1, Pos(Delimiter,working));
  Strings.add(working3{+Valueequals+working});
  end;
 if working2 <> Section then
  begin
  cSection := working2;
  if havefound then break;
  end;
 end;
end;

procedure TSupraINI.ReadSectionValues(const Section: string; Strings: TStrings);
{Read all the values from a section's keys into a string list
Modification of INI Builder function.}
var
 position : integer;
 working, Working2, working3 : string;
 CSection : string;
 havefound : boolean;
begin
cSection := ''; havefound := false;
For position := 0 to fMaindata.count-1 do
 begin
 {Get Section names}
 Working := FMainData.strings[position];
 Working2 := copy(working, 1, Pos(Delimiter,working)-1);
 delete(working, 1, Pos(Delimiter,working));
 if Working2 = Section then
  begin
  havefound := true;
  cSection := Working2;
  Working3 := copy(working, 1, Pos(Delimiter,working)-1);
  Delete(working, 1, Pos(Delimiter,working));
  Strings.add(working3+Valueequals+working);
  end;
 if working2 <> Section then
  begin
  cSection := working2;
  if havefound then break;
  end;
 end;
end;

{
------------------------------------------------------------
------------------Read/Write Functions----------------------
------------------------------------------------------------
}
function TSupraINI.ReadString(Section, Key, Vdefault : string): string;
var temp : string;
begin
temp := GetValue(Section+Delimiter+Key);
if temp = StartDefault then temp := VDefault;
Result := temp;
end;

function TSupraINI.ReadInteger(Section, Key : string; Vdefault : integer): integer;
{Read an integer value}
var
temp : string;
int : integer;
begin
temp := GetValue(Section+Delimiter+Key);
if temp = StartDefault then int := VDefault
 else int := strtoint(temp);
Result := int;
end;

function TSupraINI.ReadBool(Section, Key : string): boolean;
{Read a boolean value.
Supports YES,Y,1,TRUE, everything else is false.}
var
 temp : string;
begin
temp := GetValue(Section+Delimiter+Key);
Temp := Uppercase(temp);
if (temp = 'YES')
 or (temp = 'Y')
 or (temp = '1')
 or (temp = 'TRUE')
  then result := true
 else Result := false;
end;

function TSupraINI.ReadFloat(Section, Key : string; Vdefault : extended): extended;
{Read a floating-point value }
var
 temp : string;
 int : extended;
begin
temp := GetValue(Section+Delimiter+Key);
if temp = StartDefault then int := VDefault
 else
  try {Exception handler for converter}
  int := strtofloat(temp);
  except
  on EConvertError do int := 0;
  end;
Result := int;
end;

procedure TSupraINI.WriteString(Section, Key, value : string);
{Write a value into the database}
var
 pos : integer;
begin
Pos := GetEntryPosition(section, Key);
if Pos <> -1 then
 begin
 FMainData.strings[pos] := Section+Delimiter+key+delimiter+value;
 end
else
 begin
 FMainData.add(Section+Delimiter+key+delimiter+value);
 end;
FMainData.Sort;
end;

procedure TSupraINI.WriteInteger(Section, Key : string; value : integer);
{Write an integer value}
var
 pos : integer;
 temp : string;
begin
{Convert value}
temp := inttostr(value);
{check for and write value}
Pos := GetEntryPosition(section, Key);
if Pos <> -1 then
 begin
 FMainData.strings[pos] := Section+Delimiter+key+delimiter+temp;
 end
else
 begin
 FMainData.add(Section+Delimiter+key+delimiter+temp);
 end;
FMainData.Sort;
end;

procedure TSupraINI.WriteBool(Section, Key : string; value : boolean);
{Write a boolean value into the database}
var
 pos : integer;
 temp : string;
begin
{Convert value}
if value then temp := BoolOne
 else temp := BoolZero;
{check for and write value}
Pos := GetEntryPosition(section, Key);
if Pos <> -1 then
 begin
 FMainData.strings[pos] := Section+Delimiter+key+delimiter+temp;
 end
else
 begin
 FMainData.add(Section+Delimiter+key+delimiter+temp);
 end;
FMainData.Sort;
end;

procedure TSupraINI.WriteFloat(Section, Key : string; value : extended);
{Write a floating-point value}
var
 pos : integer;
 temp : string;
begin
{Convert value}
temp := FloatToStrF(value, ffGeneral, 18, 0);
{check for and write value}
Pos := GetEntryPosition(section, Key);
if Pos <> -1 then
 begin
 FMainData.strings[pos] := Section+Delimiter+key+delimiter+temp;
 end
else
 begin
 FMainData.add(Section+Delimiter+key+delimiter+temp);
 end;
FMainData.Sort;
end;

end.
